PeriodSPEI <- function(tas = NULL, tasmax = NULL, tasmin = NULL, pet = NULL, prlr, # s2dv_cubes (with coordinates lat = prlr$coords$latitude and Dates)
                        time_dim = 'syear', leadtime_dim = 'time', memb_dim = 'ensemble', lon_dim = 'longitude', lat_dim = 'latitude',
                        accum = 1, start = NULL, end = NULL,
                        pet_method = NULL,
                        standardization = TRUE,
                        param_error = -9999,
                        handle_infinity = FALSE,
                        cross_validation = FALSE,
                        method = 'parametric', distribution = 'log-Logistic', fit='ub-pwm',
                        n_procs = 4){

      # check provided data and parameters:
        if (!is.null(pet) & !is.null(pet_method)){
          print ('WARNING: pet data is provided and also a pet estimation method, the provided pet data will be used and NOT estimated')
          pet_method <- NULL
        }
        if (is.null(pet_method) & is.null(pet)){
          stop ('variable pet needs to be provided or a pet_method selected')
        }
        if ((pet_method == 'hargreaves' | pet_method == 'hargreaves_modified') & (is.null(tasmax) | is.null(tasmin))){
          stop (paste0('Need to provide tasmax and tasmin for the selected method ', pet_method))
        }
        if (pet_method == 'thornthwaite' & is.null(tas)){
          stop (paste0('Need to provide tas for the selected method ', pet_method))
        }
        print('WARNING: temperature needs to be in C and precipitation in mm/month') # there is no check

      # check if accumulation period is possible
        if(accum > dim(prlr$data)[leadtime_dim][[1]]){
          stop(paste0('ERROR: Cannot compute accumulation of ', accum, ' months because loaded data has only ', dim(prlr$data)[leadtime_dim][[1]], ' months.'))
        }

      # complete dates
        dates <- prlr$attrs$Dates 
        dates_complete_daily <- as.Date(as.Date(paste(lubridate::year(min(dates)), 01, 01, sep='-')):as.Date(paste(lubridate::year(max(dates)), 12, 31, sep='-')))
        dates_complete_daily_to_monthly <- lubridate::day(dates_complete_daily)
        dates_complete_monthly <- dates_complete_daily[which(dates_complete_daily_to_monthly == 1)]
        dates_monthly <- array(0, dim=length(dates_complete_monthly))
        for (dd in 1:length(dates)){
          ii <- which(dates_complete_monthly == as.Date(paste(lubridate::year(dates[dd]), lubridate::month(dates[dd]), 01, sep = '-')))
          dates_monthly[ii] <- 1
        }

      # Evapotranspiration estimation (unless pet is already provided)
        if (is.null(pet)){
          if (pet_method == 'hargreaves'){
            data <- list(tasmax = tasmax$data, tasmin = tasmin$data, lat = prlr$coords$latitude)
          } else if (pet_method == 'hargreaves_modified'){
            data <- list(tasmax = tasmax$data, tasmin = tasmin$data, prlr = prlr$data, lat = prlr$coords$latitude)
          } else if (pet_method == 'thornthwaite'){
            data <- list(tas = tas$data, lat = prlr$coords$latitude)
          }
          pet <- evapotranspiration(data, dates_monthly, pet_method, time_dim, leadtime_dim, memb_dim, lon_dim, lat_dim, n_procs)
        }

      # Accumulation
        diff_P_PET <- prlr$data - pet
        data_accum <- accumulation(diff_P_PET, dates_monthly, accum, time_dim, leadtime_dim, memb_dim, n_procs)

      # Standardization: 
        if (standardization == TRUE){
          spei_dat <- spei_standardization(data_accum = data_accum,
                                          leadtime_dim = leadtime_dim,
                                          time_dim = time_dim,
                                          memb_dim = memb_dim,
                                          cross_validation = cross_validation,
                                          handle_infinity = handle_infinity,
                                          n_procs = n_procs,
                                          accum = accum,
                                          param_error = param_error,
                                          method = method, distribution = distribution, fit = fit)
        } else {
          spei_dat <- data_accum
        }

  return(spei_dat)
}

    

## functions

evapotranspiration <- function(data, dates_monthly, pet_method = 'hargreaves', 
                                time_dim = 'syear', leadtime_dim = 'time', memb_dim = 'ensemble', 
                                lon_dim = 'longitude', lat_dim = 'latitude', 
                                n_procs = 4){

  lat_mask <- InsertDim(InsertDim(data$lat,pos = 1, len = 1, name = 'dat'),pos = 3, len = dim(data[[1]])[lon_dim], name = lon_dim)

  # extract mask of NA locations to return to NA the final result
    mask_NA <- array(1, dim = dim(data[[1]]))
    if (pet_method == 'hargreaves'){
      mask_NA[which(is.na(data$tasmax))] <- 0
      mask_NA[which(is.na(data$tasmin))] <- 0
    }
    if (pet_method == 'hargreaves_modified'){
      mask_NA[which(is.na(data$tasmax))] <- 0
      mask_NA[which(is.na(data$tasmin))] <- 0
      mask_NA[which(is.na(data$prlr))] <- 0
    }
    if (pet_method == 'thornthwaite'){
      mask_NA[which(is.na(tas$data))] <- 0
    }
  
  # replace NA with 0
  for (dd in 1:length(data)){
    data[[dd]][which(is.na(data[[dd]]))] <- 0
  }

  # prepare data
  if (pet_method == 'hargreaves'){
    data_input <- list(lat_mask = lat_mask, tasmax = data$tasmax, tasmin = data$tasmin)
    targetdim_input <- list(lat_mask = c('dat'), tasmax = c(leadtime_dim, time_dim), tasmin = c(leadtime_dim, time_dim))
  } else if (pet_method == 'hargreaves_modified'){
    data_input <- list(lat_mask = lat_mask, tasmax = data$tasmax, tasmin = data$tasmin, prlr = data$prlr)
    targetdim_input <- list(lat_mask = c('dat'), tasmax = c(leadtime_dim, time_dim), tasmin = c(leadtime_dim, time_dim), prlr = c(leadtime_dim, time_dim))
  } else if (pet_method == 'thornthwaite'){
    data_input <- list(lat_mask = lat_mask, tas = data$tas)
    targetdim_input <- list(lat_mask = c('dat'), tas = c(leadtime_dim, time_dim))
  } else {
    stop (paste0('Unknown pet_method ', pet_method))
  }

  PET_estimated <- Apply(data = data_input, 
                         target_dims = targetdim_input, 
                         output_dims = c(leadtime_dim, time_dim),
                         pet_method = pet_method, 
                         dates_monthly = dates_monthly, # array of 0 and 1 used to build the data_input to a complete array of the full years that are considered
                         dim_month = leadtime_dim, dim_year = time_dim,
                         fun = atomic_pet, ncores = n_procs)
  
  # reorder dims in PET_estimated
    dims_order <- array(NA, length(dim(data[[1]])))
    for (ord in 1:length(dim(data[[1]]))){
      dims_order[ord] <- which(names(dim(PET_estimated$output1)) == names(dim(data[[1]]))[ord])
    }
    data_pet <- aperm(PET_estimated$output1, dims_order)

  # restore original NAs from mask_NA
    data_pet[which(mask_NA == 0 )] <- NA

  return(data_pet)

}

accumulation <- function(diff_P_PET, dates_monthly, accum = 1, time_dim = 'syear', leadtime_dim = 'time', memb_dim = 'ensemble', n_procs = 4){
  
  if(!time_dim %in% names(dim(diff_P_PET))){
    diff_P_PET <- InsertDim(diff_P_PET, posdim = 1, lendim = 1, name = time_dim)
  }
  if(!leadtime_dim %in% names(dim(diff_P_PET))){
    diff_P_PET <- InsertDim(diff_P_PET, posdim = 1, lendim = 1, name = leadtime_dim)
  }
  
  accum_result <- Apply(data = list(diff_P_PET), 
                        target_dims = list(diff_P_PET = c(leadtime_dim,time_dim)), 
                        dates_monthly = dates_monthly, # array of 0 and 1 used to build the data_input to a complete array of the full years that are considered
                        accum = accum,
                        output_dims = c(leadtime_dim, time_dim), #c('time','sdate'),
                        dim_month = leadtime_dim, dim_year = time_dim,
                        fun = atomic_accum, ncores = n_procs)
    
  # recover essential lost dims (if they had length 1 they'd have been dropped in previous step):
  for (d in c(time_dim, leadtime_dim, 'latitude', 'longitude', memb_dim)){
    if(!d %in% names(dim(accum_result$output1))){ 
      accum_result$output1 <- InsertDim(data = accum_result$output1, posdim = length(names(dim(accum_result$output1))) + 1, lendim = 1, name = d)
    }
  }
  
  # reorder dims in accum_result
    dims_order <- array(NA, length(dim(diff_P_PET)))
    for (ord in 1:length(dim(diff_P_PET))){
      dims_order[ord] <- which(names(dim(accum_result$output1)) == names(dim(diff_P_PET))[ord])
    }
    data_accum <- aperm(accum_result$output1, dims_order)

  return(data_accum)

}

spei_standardization <- function(data_accum,
                                 leadtime_dim, time_dim, memb_dim, handle_infinity, cross_validation,
                                 n_procs, accum, param_error,
                                 method ='parametric', distribution = 'log-Logistic', fit='ub-pwm'){
  data_spei <- data_accum

  data_spei <- Apply(data = list(data_accum), 
                      target_dims = list(data = c(leadtime_dim,time_dim, memb_dim)),
                      output_dims = c(leadtime_dim,time_dim, memb_dim), 
                      dim_month = leadtime_dim, 
                      dim_year = time_dim,
                      dim_memb = memb_dim,
                      handle_infinity = handle_infinity,
                      cross_validation = cross_validation,
                      method = method, distribution = distribution, fit = fit, 
                      ref_period = NULL,
                      param_error = param_error,
                      fun = atomic_spei, 
                      ncores = n_procs)$output1
  
  return(data_spei)

}

### Atomic functions

atomic_pet <- function(pet_method, dates_monthly, dim_month, dim_year, lat_mask = NULL, data2 = NULL, data3 = NULL, data4 = NULL){
 
  ftime = as.numeric(dim(data2)[leadtime_dim])
  n_sdates = as.numeric(dim(data2)[time_dim])

  # create a vector from data but adding 0 to achive complete time series of the considered period 
    #(starting in January of the first year) so that the solar radiation estimation is computed 
    # in each case for the correct month
    if (!is.null(data2)){
      data_tmp <- as.vector(data2)
      data2 <- array(0, dim = length(dates_monthly))
      count <- 1
      for (dd in 1:length(dates_monthly)){
        if (dates_monthly[dd] == 1){
          data2[dd] <- data_tmp[count]
          count <- count + 1
        }
      }
      rm(data_tmp)
    }
    if (!is.null(data3)){
      data_tmp <- as.vector(data3)
      data3 <- array(0, dim = length(dates_monthly))
      count <- 1
      for (dd in 1:length(dates_monthly)){
        if (dates_monthly[dd] == 1){
          data3[dd] <- data_tmp[count]
          count <- count + 1
        }
      }
      rm(data_tmp)
    }
    if (!is.null(data4)){
      data_tmp <- as.vector(data4)
      data4 <- array(0, dim = length(dates_monthly))
      count <- 1
      for (dd in 1:length(dates_monthly)){
        if (dates_monthly[dd] == 1){
          data4[dd] <- data_tmp[count]
          count <- count + 1
        }
      }
      rm(data_tmp)
    }
  if (pet_method == 'hargreaves'){
    x_PET <- hargreaves(Tmin = as.vector(data3), Tmax = as.vector(data2), lat = lat_mask, na.rm = FALSE)
    x_PET <- x_PET[which(dates_monthly == 1)] # line to return the vector to the size of the actual original data
    PET <- array(x_PET, dim = c(ftime,n_sdates))
  }

  if (pet_method == 'hargreaves_modified'){ # not sure this works properly
    PET <- array(NA, dim = c(ftime,n_sdates))
    for(ns in 1:n_sdates){
      tmax_mod_temp <- data2[12*(ns-1)+(1:12)] #data2[,ns] 
      tmin_mod_temp <- data3[12*(ns-1)+(1:12)] #data3[,ns]
      pre_mod_temp <- data4[12*(ns-1)+(1:12)] #data4[,ns]
      
      # Computation of PET
      x_PET <- hargreaves(Tmin = tmin_mod_temp, Tmax = tmax_mod_temp, lat = lat_mask, Pre = pre_mod_temp, na.rm = FALSE)
      PET[1:ftime,ns] <- x_PET      #dim <- (lon lat ft sd memb)                
    }
  }
  
  if (pet_method == 'thornthwaite'){
    x_PET <- thornthwaite(as.vector(data2), lat = lat_mask, na.rm = TRUE)
    x_PET <- x_PET[which(dates_monthly == 1)] # line to return the vector to the size of the actual original data
    PET <- array(x_PET, dim = c(ftime,n_sdates))
  }
  
  return(PET)
  
}

atomic_accum <- function(data, dates_monthly, accum, dim_month, dim_year){
  # data:[time, syear]

  ftime = as.numeric(dim(data)[dim_month])
  n_sdates = as.numeric(dim(data)[dim_year])
  
  data_vector <- array(0, dim = length(dates_monthly))
  count <- 1
  for (dd in 1:length(dates_monthly)){
    if (dates_monthly[dd] == 1){
      data_vector[dd] <- as.vector(data)[count]
      count <- count + 1
    }
  }

  # Accumulation at different timescales
  data_sum_x <- rollapply(data_vector, accum, sum) # rollapply {zoo} A generic function for applying a function to rolling margins of an array.
  data_sum_x <- c(rep(NA, accum-1), data_sum_x) # adds as many NAs as needed at the begining to account for the months that cannot be added (depends on accu) and so that the position in the vector corresponds to the accumulated of the previous months (instead of the accumulated of the next months)
  data_sum_x <- data_sum_x[which(dates_monthly == 1)] # discard the months that don't appear in the original data
  accum_result <- array(data_sum_x, dim = c(ftime,n_sdates)) # return to matrix form
  if (accum > 1){
    accum_result[1:(accum-1),] <- NA # replace by NA when the accumulation corresponds to months that where not present in the original data
  }

  return(accum_result)
  
}

atomic_spei <- function(data, dim_month, dim_year, dim_memb, ref_period, handle_infinity, cross_validation, param_error, 
                          method ='parametric', distribution = 'log-Logistic', fit='ub-pwm'){
  # data: [time, sdate, memb]
  if (is.null(ref_period)){
    ref.start <- NULL
    ref.end <- NULL
  } else {
    ref.start <- ref_period[[1]]
    ref.end <- ref_period[[2]]
  }

  if (all(is.na(data))) {
    speiX <- array(NA, dim(data))
  } else if (var(data, na.rm=T) == 0) { # if the data [time, sdate, memb] has no variability it will raise an error further down, so we assign a value to the result and skip the step
    speiX <- array(param_error, dim(data))
  } else {
    speiX <- spX(data, ref.start=ref.start, ref.end=ref.end, 
                  method = method, distribution = distribution, fit = fit, 
                  dim_month = dim_month, dim_year = dim_year, dim_memb = dim_memb, 
                  handle_infinity = handle_infinity, cross_validation = cross_validation)
  }
  return(speiX)
  
}

### Support functions

spX <- function(acc_data, ref.start, ref.end, na.rm = TRUE, method = 'parametric', distribution = 'log-Logistic', fit = 'ub-pwm', dim_month, dim_year, dim_memb, handle_infinity, cross_validation){
  
  # acc_data: [ftime, syear, nmemb]

  if (!(method %in% c('parametric', 'non-parametric'))) {
    stop('pcX script error: SPEI can be only computed using the following approach: parametric or non-parametric')
  }
  
  ftime <- as.numeric(dim(acc_data)[dim_month])
  n_sdates <- as.numeric(dim(acc_data)[dim_year])
  nmemb <- as.numeric(dim(acc_data)[dim_memb])

  spei_mod <- array(data = NA, dim = c(ftime, n_sdates, nmemb))
  names(dim(spei_mod)) <- c(dim_month, dim_year, dim_memb)

  for(ff in 1:ftime){ # treat months independently   
    Accum_D_temp <- ClimProjDiags::Subset(acc_data, along = dim_month, indices = ff, drop = FALSE)
    x_spei_mod <- spX_ftime(Accum_D_temp, na.rm = TRUE, method = method, distribution = distribution, fit = fit, ref.start = ref.start, ref.end = ref.end, n_sdates = n_sdates, nmemb = nmemb, handle_infinity = handle_infinity, cross_validation = cross_validation)
    spei_mod[ff,,] <- x_spei_mod
  }

  return(spei_mod)
}

# HandleInfinity:
spX_ftime <- function(data, distribution = 'log-Logistic', fit = 'ub-pwm', na.rm = TRUE, method = 'parametric', n_sdates, nmemb,
                             ref.start = NULL, ref.end = NULL, params = NULL, handle_infinity = FALSE, cross_validation = FALSE) {
  
  # data:[ftime = 1, syear, ensemble]

  if (!(method %in% c('parametric', 'non-parametric'))) {
    stop('SPEI can be only computed using the following approach: parametric or non-parametric')
  }
  
  if (method == 'non-parametric'){ 
    
    if (anyNA(data) && na.rm == FALSE) {
      stop('Error: Data must not contain NAs')
    }
    
    bp = matrix(0, length(data), 1)
    for (i in 1:length(data)) {
      bp[i,1] = sum(data[] <= data[i]);  # Writes the rank of the data
    }
    
    SPEI = qnorm((bp-0.44)/(length(data)+0.12))
    
    return(SPEI)
    
  } else {

    std_index <- array(NA, c(n_sdates, nmemb))
  
    if (anyNA(data) && na.rm==FALSE) {
      stop('Error: Data must not contain NAs')
    }
    if (!(distribution %in% c('log-Logistic', 'Gamma', 'PearsonIII'))) {
      stop('Distrib must be one of "log-Logistic", "Gamma" or "PearsonIII"')
    }
    if (!(fit %in% c('max-lik', 'ub-pwm', 'pp-pwm'))) {
      stop('Method must be one of "ub-pwm" (default), "pp-pwm" or "max-lik"')
    }
    
    coef = switch(distribution,
                  "Gamma" = array(NA, dim = 2, dimnames = list(c('alpha','beta'))),
                  "log-Logistic" = array(NA, dim = 3, dimnames = list(c('xi','alpha','kappa'))),
                  "PearsonIII" = array(NA, dim = 3, dimnames = list(c('mu','sigma','gamma')))
    )

    dim_one <- length(coef)

    if (!is.null(params)) {
      if (length(params)!=dim_one) {
        stop(paste0('parameters array should have dimensions [', dim_one, ']'))
      }
    }
    
    # Select window if necessary
      if (!is.null(ref.start) && !is.null(ref.end)) {
        data.fit <- window(data,ref.start,ref.end)	
      } else {
        data.fit <- data
      }

    if(cross_validation == "TRUE") {  
      loop_years <- n_sdates
    } else {
      loop_years <- 1
    }
    
    for (nsd in 1:loop_years) { # loop over years (in case of cross_validation)
      # Cumulative series (acu)
      if (cross_validation == TRUE){
        acu <- as.vector(data.fit[,-nsd,])
      } else {
        acu <- as.vector(data.fit)
      }
        
      acu.sorted <- sort.default(acu, method = "quick")
      acu.sorted <- acu.sorted[!is.na(acu.sorted)] # remove NAs (no need if(na.rm) because if there are NA and na.rm=F we don't get to this point)
      if (length(acu.sorted)!=0){ # else all acu was NA and we don't need to continue with this case     
        acu_sd = sd(acu.sorted)
        if (!is.na(acu_sd)){
        if (acu_sd != 0){
          if(distribution != "log-Logistic"){
            pze <- sum(acu==0)/length(acu)
            acu.sorted = acu.sorted[acu.sorted > 0]
          }
          if (!is.null(params)) {
            f_params = as.vector(params)
          } else {
            if (length(acu.sorted) >= 4){ # else coef will be NA
              # Calculate probability weighted moments based on fit with lmomco or TLMoments
              pwm = switch(fit,
                            "pp-pwm" = pwm.pp(acu.sorted,-0.35,0, nmom=3),
                            pwm.ub(acu.sorted, nmom=3)
                            #TLMoments::PWM(acu.sorted, order=0:2)
              )

              # Check L-moments validity
              lmom <- pwm2lmom(pwm)          
              if (!(!are.lmom.valid(lmom) || anyNA(lmom[[1]]) || any(is.nan(lmom[[1]])))){
                
                # lmom fortran functions need specific inputs L1, L2, T3
                # this is handled by lmomco internally with lmorph
                fortran_vec = c(lmom$lambdas[1:2], lmom$ratios[3])

                # Calculate parameters based on distribution with lmom then lmomco
                f_params = switch(distribution,
                          "log-Logistic" = tryCatch(lmom::pelglo(fortran_vec), error = function(e){ parglo(lmom)$para }),
                          "Gamma" = tryCatch(lmom::pelgam(fortran_vec), error = function(e){ pargam(lmom)$para }),
                          "PearsonIII" = tryCatch(lmom::pelpe3(fortran_vec), error = function(e){ parpe3(lmom)$para })
                )

                # Adjust if user chose log-Logistic and max-lik
                if(distribution == 'log-Logistic' && fit == 'max-lik'){
                  f_params = parglo.maxlik(acu.sorted, f_params)$para
                }
              } # end if dor the case the L-moments are not valid (std_index will be NA)
            } # end if for the case there are not enough values to estimate the parameters (std_index will be NA)
          } # end estimation of f_param
          # Calculate cdf based on distribution with lmom
          cdf_res = switch(distribution,
                            "log-Logistic" = lmom::cdfglo(data, f_params),
                            "Gamma" = lmom::cdfgam(data, f_params),
                            "PearsonIII" = lmom::cdfpe3(data, f_params)				  				
          )
          
          std_index_cv <- array(qnorm(cdf_res), dim = c(n_sdates,nmemb))
  
          # Adjust if user chose Gamma or PearsonIII - Not tested: For future development
          #if(distribution != 'log-Logistic'){ 
          #  std_index[ff,s] = qnorm(pze + (1-pze)*pnorm(std_index[ff,s])) # ff doesn't exist at this point
          #}
          if (cross_validation == TRUE){
            std_index[nsd,] <- std_index_cv[nsd,]  
          } else {
            std_index <- std_index_cv
          }          
        }
        } # end if for the case there is no variability        
      } # end if for the case all NA in acu
    } # next year (in case of cross_validation or all done if cross_validation == F)
            
    if(handle_infinity == 'TRUE'){ # could also use "param_error" ?; we are giving it the min/max value of the grid point
      std_index[is.infinite(std_index) & std_index < 0] <- min(std_index[!is.infinite(std_index)])
      std_index[is.infinite(std_index) & std_index > 0] <- max(std_index[!is.infinite(std_index)]) 
    }
      
  } # end parametric

  return(std_index) # fitted is the only thing used after calling this function

}

